"""This submodule defines the main window of our application."""
from PySide6 import QtWidgets
from PySide6.QtCore import Qt, QThreadPool
from PySide6.QtGui import QCloseEvent, QKeyEvent

from melon.melon import Melon
from melon.todo import Todo

from .calendarlist import CalendarListView
from .tasklist import TaskListView, UserRole


class GuiMelon(Melon):
    """Subclasses the `Melon` main object, adding a handler for changes within the data."""

    def __init__(self) -> None:
        """Initialises the Melon."""
        super().__init__()
        self.tasklistView: TaskListView | None = None

    def addOrUpdateTask(self, todo: Todo):
        """
        Args:
            todo (Todo): Argument
        """
        assert self.tasklistView is not None
        uid = todo.uid
        assert uid is not None
        for row in range(self.tasklistView.count()):
            data = self.tasklistView.item(row).data(UserRole)
            if getattr(data, "uid", None) == uid:
                if todo.isComplete():
                    self.tasklistView.takeItem(row)
                else:
                    self.tasklistView.blockSignals(True)
                    self.tasklistView.item(row).setText(todo.summary)
                    self.tasklistView.item(row).setData(UserRole, todo)
                    self.tasklistView.blockSignals(False)
                return
        if todo.isIncomplete():
            self.tasklistView.addTask(todo)


class MainWindow(QtWidgets.QWidget):
    """Main Window class that defines most of the UI."""

    def __init__(self) -> None:
        """Initialises the main window."""
        super().__init__()
        self.melon = GuiMelon()
        self.threadPool = QThreadPool()
        self.setWindowTitle("Melon UI")

    def buildUI(self):
        """
        Args:
        """
        self.tasklistView = TaskListView(self.melon)
        self.melon.tasklistView = self.tasklistView
        self.calendarlistView = CalendarListView()
        self.calendarlistView.currentItemChanged.connect(self.calendarListClicked)
        self.searchWidget = QtWidgets.QLineEdit()
        self.searchWidget.setPlaceholderText("Search tasks...")
        self.messageLabel = QtWidgets.QLabel(self)
        self.messageLabel.setHidden(True)
        self.messageLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)

        layout = QtWidgets.QGridLayout(self)
        layout.addWidget(self.messageLabel, 0, 1, 1, 2)
        # layout.addWidget(self.searchWidget, 0, 1, 1, 2)
        layout.addWidget(self.calendarlistView, 1, 1)
        layout.addWidget(self.tasklistView, 1, 2)
        self.setLayout(layout)

    def start(self):
        """
        Args:
        """
        self.melon.autoInit()
        self.tasklistView.sortItems()
        self.calendarlistView.populate(self.melon.calendars.values())
        # QTimer.singleShot(200, self.sync)

    def sync(self):
        """
        Args:
        """
        self.showInfoMessage("Syncing...")
        QtWidgets.QApplication.processEvents()
        self.melon.syncAll()
        self.tasklistView.sortItems()
        self.hideMessage()

    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Args:
            event (QCloseEvent): Argument

        """
        self.melon.store()
        return super().closeEvent(event)

    def calendarListClicked(self, item: QtWidgets.QListWidgetItem):
        """
        Args:
            item (QListWidgetItem): Argument
        """
        userData = item.data(Qt.ItemDataRole.UserRole)
        if userData and userData["is-special"] and userData["specialty"] == "all":
            self.tasklistView.clearCalendarFilter()
        else:
            self.tasklistView.setCalendarFilter(item.text())

    def keyPressEvent(self, event: QKeyEvent):
        """
        Args:
            event (QKeyEvent): Argument
        """
        if event.modifiers() == Qt.KeyboardModifier.ControlModifier:
            if event.key() == Qt.Key.Key_W:
                self.close()
            elif event.key() == Qt.Key.Key_S:
                self.sync()
            elif event.key() == Qt.Key.Key_H:
                self.calendarlistView.setCurrentRow(0)
            elif event.key() == Qt.Key.Key_Plus:
                self.tasklistView.addEmptyTask()
            elif event.key() == Qt.Key.Key_Return:
                self.threadPool.start(lambda: self.melon.scheduleAllAndExport("task-schedule.ics"))
        return super().keyPressEvent(event)

    def showInfoMessage(self, msg: str):
        """
        Args:
            msg (str): Argument
        """
        self.messageLabel.setText(msg)
        self.messageLabel.setHidden(False)

    def hideMessage(self):
        """
        Args:
        """
        self.messageLabel.setHidden(True)



"""This file is the main entry point of the melon package, containing the Melon class,
the main point of contact for users of this package. It can be initialised like this:

melon = Melon()
melon.autoInit()
"""
import datetime
import json
import logging
import uuid
from typing import Iterable, Mapping

import caldav
import caldav.lib.url
import icalendar

from .calendar import Calendar, Syncable
from .config import CONFIG, CONFIG_FOLDER
from .scheduler.base import AbstractScheduler, Task, TimeSlot
from .scheduler.purepython import MCMCScheduler
from .todo import Todo


class Melon:
    """The Melon class, wrapping a caldav client and principal, loading specifics from the config.
    Through me, users have access to calendars and todos.
    I also handle load, sync and store functionality.
    """

    HIDDEN_CALENDARS = ("calendar", None)

    def __init__(
        self,
        url=CONFIG["client"]["url"],
        username=CONFIG["client"]["username"],
        password=CONFIG["client"]["password"],
        maxCalendars: int | None = None,
    ) -> None:
        """Initialises the Melon client

        Args:
            url (str, optional): URL to the CalDAV server. Defaults to CONFIG["client"]["url"].
            username (str, optional): Username. Defaults to CONFIG["client"]["username"].
            password (str, optional): Password. Defaults to CONFIG["client"]["password"].
            maxCalendars (int, optional): the highest number of calendars to load. Useful for testing.
        """
        self.client = caldav.DAVClient(url=url, username=username, password=password)
        self.calendars: dict[str, Calendar] = {}
        self.principal = None
        self.maxCalendars: int | None = maxCalendars

    def connect(self):
        """
        Args:
        """
        self.principal = self.client.principal()
        logging.info("Obtained principal")

        all_calendars = self.principal.calendars()
        if self.maxCalendars is not None:
            all_calendars = all_calendars[: self.maxCalendars]
        self.calendars = {cal.name: Calendar(cal) for cal in all_calendars if cal.name not in self.HIDDEN_CALENDARS}
        logging.info(f"Obtained {len(self.calendars)} calendars")

    def _load_syncable_tasks(self, calendar):
        """
        Args:
            calendar: Argument
        """
        for object in calendar.syncable:
            if "vtodo" in object.vobject_instance.contents:
                assert isinstance(object, Todo)
                self.addOrUpdateTask(object)

    def fetch(self):
        """
        Args:
        """
        if not self.calendars:
            self.connect()
        for calendar in self.calendars.values():
            assert calendar.name is not None
            calendar.syncable = Syncable.upgrade(calendar.objects_by_sync_token(load_objects=True), calendar.name)
            self._load_syncable_tasks(calendar)
            logging.info(f"Fetched {len(calendar.syncable)} full objects!")

    def store(self):
        """
        Args:
        """
        for calendar in self.calendars.values():
            calendar.storeToFile()
        with open(CONFIG_FOLDER / "synctokens.json", "w") as f:
            json.dump({cal.name: cal.storageObject() for cal in self.calendars.values()}, f)
        logging.info(f"Stored {len(self.calendars)} calendars to disk.")

    def load(self):
        """
        Args:
        """
        if self.principal is None:
            self.principal = self.client.principal()
            logging.info("Obtained principal")
        with open(CONFIG_FOLDER / "synctokens.json") as f:
            data = json.load(f)
        for file in CONFIG_FOLDER.glob("*.dav"):
            name = file.stem  # filename corresponds to the calendar name
            self.calendars[name] = Calendar.loadFromFile(
                self.client, self.principal, name, data[name]["token"], data[name]["url"]
            )
            if self.maxCalendars is not None and len(self.calendars) >= self.maxCalendars:
                break
        logging.info(f"Loaded {len(self.calendars)} calendars from disk.")
        for calendar in self.calendars.values():
            self._load_syncable_tasks(calendar)

    def autoInit(self):
        """
        Args:
        """
        tokensfile = CONFIG_FOLDER / "synctokens.json"
        if not tokensfile.exists():
            self.fetch()
        else:
            self.load()

    def syncCalendar(self, calendar: Calendar):
        """
        Args:
            calendar: Argument
        """
        calendar.sync()
        self._load_syncable_tasks(calendar)

    def syncAll(self):
        """
        Args:
        """
        for calendar in self.calendars.values():
            self.syncCalendar(calendar)

    def allTasks(self) -> Iterable[Todo]:
        """Returns an iterable of all tasks in all calendars as a single list

        Yields:
            Iterator[Iterable[Todo]]: iterator of all tasks
        """
        for calendar in self.calendars.values():
            if calendar.syncable is None:
                continue
            for object in calendar.syncable.objects:
                if object.uid is None or not object.isTodo():
                    continue
                yield object

    def allIncompleteTasks(self) -> Iterable[Todo]:
        """Returns all incomplete todos

        Yields:
            Iterator[Iterable[Todo]]: incomplete todos
        """
        for todo in self.allTasks():
            if todo.isIncomplete():
                yield todo

    def getTask(self, uid: str) -> Todo:
        """Returns task with given UID

        Args:
            uid (str): the Unique Identifier

        Raises:
            ValueError: when the task could not be found

        Returns:
            Todo: the Todo with given uid
        """
        for object in self.allTasks():
            if object.uid == uid:
                return object
        raise ValueError(f"Task with UID {uid} not found.")

    def findTask(self, string: str) -> Iterable[Todo]:
        """Finds a task given a search query

        Args:
            string (str): the search query.

        Yields:
            Iterator[Iterable[Todo]]: the generated search results.
        """
        for object in self.allTasks():
            if string in object.data and object.isTodo():
                yield object

    def addOrUpdateTask(self, todo: Todo):
        """
        Args:
            todo (Todo): Argument
        """

    def exportScheduleAsCalendar(self, scheduling: Mapping[str, TimeSlot]) -> icalendar.Calendar:
        """A read-only ICS calendar containing scheduled tasks. Can be stored to disk using schedule.to_ical().

        Args:
            scheduling (Mapping[str, TimeSlot]): Mapping of task UID to TimeSlot

        Returns:
            icalendar.Calendar: the calendar containing events (time slots) proposed for the completion of tasks
        """
        schedule = icalendar.Calendar()
        schedule.add("prodid", "-//Melon//example.org//")
        schedule.add("version", "2.0")
        todos = {t.uid: t for t in self.allTasks()}
        for uid, slot in scheduling.items():
            event = icalendar.Event(summary=todos[uid].summary)
            event.add("dtstart", slot.timestamp)
            event.add("dtend", slot.end)
            event.add("dtstamp", datetime.datetime.now())
            event.add("uid", uuid.uuid4())
            schedule.add_component(event)
        return schedule

    def tasksToSchedule(self) -> list[Task]:
        """Returns all incomplete tasks as scheduler.Task objects

        Returns:
            list[Task]: _description_
        """
        return list(map(Todo.toTask, self.allIncompleteTasks()))

    def scheduleAllAndExport(self, file: str, Scheduler: type[AbstractScheduler] = MCMCScheduler):
        """Runs the scheduler on all tasks and exports as an ICS file.

        Args:
            file (str): filesystem path that the ics file should be exported to
        """
        logging.info("Initialising scheduler.")
        scheduler = Scheduler(self.tasksToSchedule())
        logging.info(f"Scheduling {len(scheduler.tasks)} tasks now.")
        schedule = scheduler.schedule()
        logging.info("Exporting.")
        export = self.exportScheduleAsCalendar(schedule)
        logging.info("Export calendar created.")
        # print(export.to_ical().decode().replace("\r\n", "\n"))
        with open(file, "wb") as f:
            f.write(export.to_ical())
        logging.info("Finished export.")
        return scheduler



"""The scheduler algorithm"""
import logging
import pathlib
import sys
from datetime import date, datetime, timedelta
from typing import Mapping

try:
    from melon.scheduler import libcppscheduler
except ImportError:
    importPath = str(pathlib.Path(__file__).resolve().parent.parent.parent / "build")
    logging.info(f"Could not find packaged .so file, falling back to {importPath} directory")
    sys.path.append(importPath)
    import libcppscheduler

from .base import START_OF_DAY, AbstractScheduler, TimeSlot


class CppMCMCScheduler(AbstractScheduler):
    """Markov Chain Monte-Carlo Task Scheduler, implemented in Rust."""

    def schedule(self) -> Mapping[str, TimeSlot]:
        """Runs the Rust implementation of the scheduler.

        Returns:
            Mapping[str, TimeSlot]: the resulting schedule
        """
        start = datetime.combine(date.today(), START_OF_DAY)  # equivalent to t = 0 for libcppscheduler
        result = libcppscheduler.schedule([task.asTuple(start) for task in self.tasks])
        return {t[0]: TimeSlot(start + timedelta(hours=t[1]), t[2]) for t in result}



"""This module contains the Todo class."""
import datetime
import logging
import re
from typing import Literal

import caldav
import caldav.lib.url
import icalendar.cal
import icalendar.prop
import vobject

from melon.scheduler.base import Task

NEW_TASK_TEXT = "An exciting new task!"
MIDNIGHT = datetime.time(0, 0)


class Todo(caldav.Todo):
    """A class representing todos (= tasks), subclassing the caldav.Todo object which in turn stores VTODO data."""

    vobject_instance: vobject.base.Component  # fast access
    icalendar_component: icalendar.cal.Todo  # slow access, but does more checks

    def __init__(self, *args, calendarName: str | None = None, **kwargs):
        """Initialises the base class"""
        super().__init__(*args, **kwargs)
        self.calendarName = calendarName

    @staticmethod
    def upgrade(todo: caldav.Todo, calendarName: str) -> "Todo":
        """A copy constructor constructing a melon.Todo from a caldav.Todo

        Args:
            todo (caldav.Todo): Argument
            calendarName (str): Argument
        """
        return Todo(todo.client, todo.url, todo.data, todo.parent, todo.id, todo.props, calendarName=calendarName)

    @property
    def vtodo(self) -> vobject.base.Component:
        """Returns the VTODO object stored within me. This is faster than accessing the icalendar_component.

        Returns:
            (vobject.base.Component):
        """
        return self.vobject_instance.contents["vtodo"][0]  # type: ignore

    @property
    def summary(self) -> str:
        """
        Returns:
            (str):
        """
        return self.vtodo.contents["summary"][0].value  # type: ignore

    @summary.setter
    def summary(self, value: str):
        """
        Args:
            value (str): Argument
        """
        self.vtodo.contents["summary"][0].value = value  # type: ignore

    @property
    def dueDate(self) -> datetime.date | None:
        """
        Returns:
            (datetime.date | None):
        """
        if "due" in self.vtodo.contents:
            due = self.vtodo.contents["due"][0].value  # type: ignore
            if isinstance(due, datetime.datetime):
                return due.date()
            return due  # otherwise, this value is already a datetime.date

    @dueDate.setter
    def dueDate(self, value: datetime.date | None) -> None:
        """Sets my due date.

        Args:
            value (datetime.date): Date to set.
        """
        if value is None:
            del self.icalendar_component["due"]
        else:
            self.icalendar_component["due"] = icalendar.prop.vDDDTypes(value)

    @property
    def dueTime(self) -> datetime.time | None:
        """
        Returns:
            (datetime.time | None):
        """
        if "due" in self.vtodo.contents:
            due = self.vtodo.contents["due"][0].value  # type: ignore
            if isinstance(due, datetime.datetime):
                return due.time()

    @property
    def dueDateTime(self) -> datetime.datetime | None:
        """
        Returns:
            (datetime.datetime | None):
        """
        if "due" in self.vtodo.contents:
            due = self.vtodo.contents["due"][0].value  # type: ignore
            if isinstance(due, datetime.date):
                return datetime.datetime.combine(due, datetime.time())
            return due

    @property
    def uid(self) -> str | None:
        """This method has to be fast, as it is accessed very frequently according to profiler output.
        Therefore we use do not use self.vtodo.

        Returns:
            (Union[str, None]):
        """
        try:
            return self._vobject_instance.contents["vtodo"][0].contents["uid"][0].value  # type: ignore
        except AttributeError:
            return self.vtodo.contents["uid"][0].value  # type: ignore
        except KeyError:
            return

    @property
    def priority(self) -> int:
        """
        Returns:
            int: the priority of the task, an integer between 1 and 9,
                 where 1 corresponds to the highest and 9 to the lowest priority
        """
        value = self.vtodo.contents.get("priority")
        return int(value[0].value) if value is not None else 9  # type: ignore

    def isIncomplete(self) -> bool:
        """
        Returns:
            (bool):
        """
        data = self.data
        return "STATUS:NEEDS-ACTION" in data or (
            not "\nCOMPLETED:" in data and not "\nSTATUS:COMPLETED" in data and not "\nSTATUS:CANCELLED" in data
        )

    def isComplete(self) -> bool:
        """
        Returns:
            (bool):
        """
        return not self.isIncomplete()

    def complete(
        self,
        completion_timestamp: datetime.datetime | None = None,
        handle_rrule: bool = True,
        rrule_mode: Literal["safe", "this_and_future"] = "safe",
    ) -> None:
        """
        Args:
            completion_timestamp (Union[datetime.datetime, None], optional): Argument
                (default is None)
            handle_rrule (bool, optional): Argument
                (default is True)
            rrule_mode (Literal['safe', 'this_and_future'], optional): Argument
                (default is 'safe')

        """
        super().complete(completion_timestamp, handle_rrule, rrule_mode)
        logging.info("Task completed.")

    def isTodo(self) -> bool:
        """
        Returns:
            bool: whether this object is a VTODO or not (i.e. an event or journal).
        """
        return "vtodo" in self.vobject_instance.contents

    def toTask(self) -> Task:
        """Converts this Todo into the scheduler-compatible Task struct.

        Returns:
            Task: a melon.scheduler.Task
        """
        assert self.uid is not None
        location = 0
        if "home" in self.summary:
            location = 1
        elif "work" in self.summary:
            location = 2
        match = re.search(r"\b([\d\,\.])+h\b", self.summary)
        hours = float(match.group(1)) if match else 1.0
        return Task(self.uid, hours, self.priority, location, self.dueDateTime)

    def __lt__(self, other: "Todo") -> bool:
        """Compares two todos in terms of ordering

        Args:
            other (Todo): the instance to compare with

        Returns:
            bool: whether self < other
        """
        if self.summary == NEW_TASK_TEXT:
            return False
        if other.summary == NEW_TASK_TEXT:
            return True
        if self.dueDate is None and other.dueDate is not None:
            return False
        if other.dueDate is None and self.dueDate is not None:
            return True
        return (
            self.dueDate,
            self.dueTime or MIDNIGHT,
            self.summary,
        ) < (
            other.dueDate,
            other.dueTime or MIDNIGHT,
            other.summary,
        )

    def __str__(self) -> str:
        """
        Returns:
            (str):
        """
        return self.summary

    def __repr__(self) -> str:
        """
        Returns:
            (str):
        """
        return f"<Melon.Todo: {self.summary}>"



"""This module contains the Calendar class."""
import logging
from typing import Iterable

import caldav
import caldav.lib.url
import icalendar
import vobject

from .config import CONFIG_FOLDER
from .todo import NEW_TASK_TEXT, Todo


class Calendar(caldav.Calendar):
    """Class representing a calendar (or todo list, if you want to call it that, this name is given by CalDAV).
    A calendar is a collection of objects that can be synced to a CalDAV server.
    In this implementation, the objects are stored within the `syncable` subclass.
    """

    def __init__(self, calendar: caldav.Calendar) -> None:
        """A copy constructor

        Args:
            calendar (caldav.Calendar): Argument
        """
        super().__init__(
            calendar.client,
            calendar.url,
            calendar.parent,
            calendar.name,
            calendar.id,
            calendar.props,
            **calendar.extra_init_options,
        )
        self.syncable: Syncable | None = None

    def storeToFile(self):
        """Save the calendar objects to a local file on disk, in iCal format."""
        ical = vobject.iCalendar()
        assert self.syncable is not None
        for task in self.syncable:
            ical.add(task.vobject_instance)
        with open(CONFIG_FOLDER / f"{self.name}.dav", "w") as f:
            ical.serialize(f)  # type: ignore

    @staticmethod
    def loadFromFile(client: caldav.DAVClient, principal: caldav.Principal, name: str, sync_token: str, url: str):
        """
        Args:
            client (caldav.DAVClient): Argument
            principal (caldav.Principal): Argument
            name (str): Argument
            sync_token (str): Argument
            url (str): Argument
        """
        cal_url = caldav.lib.url.URL(url)
        with open(CONFIG_FOLDER / f"{name}.dav") as f:
            ical = icalendar.Calendar.from_ical(f.read())
        objects = []
        cal = Calendar(caldav.Calendar(client, parent=principal.calendar_home_set, name=name, url=url))
        for task in ical.subcomponents:
            todo = Todo.upgrade(caldav.Todo(client, parent=cal), name)
            todo.icalendar_instance = task
            todo.url = cal_url.join(str(task.subcomponents[0].get("uid")) + ".ics")
            objects.append(todo)
        cal.syncable = Syncable(cal, objects, sync_token)
        logging.info(f"Calendar {name}: loaded {len(objects)} objects.")
        return cal

    def createTodo(self, summary: str = NEW_TASK_TEXT):
        """
        Args:
            summary (str): Argument
        """
        assert self.name is not None
        return Todo(
            self.client,
            data=self._use_or_create_ics(f"SUMMARY:{summary}", objtype="VTODO"),  # type: ignore
            parent=self,
            calendarName=self.name,
        )

    def storageObject(self) -> dict:
        """
        Returns:
            (dict):
        """
        assert self.syncable is not None
        return {
            "url": str(self.url),
            "token": self.syncable.sync_token,
        }

    def sync(self):
        """Synchronise me"""
        assert self.name is not None
        assert self.syncable is not None
        updated, deleted = self.syncable.sync()
        self.syncable.upgradeObjects(self.name)
        logging.info(
            f"Synced {self.name:48} ({len(updated)} updated and {len(deleted)} deleted entries.) "
            f"In total, we have {len(self.syncable)} objects."
        )


class Syncable(caldav.SynchronizableCalendarObjectCollection):
    """The synchronisable collection of CalDAV objects, handling efficient syncs between server and client."""

    calendar: Calendar
    objects: Iterable[Todo]
    sync_token: str

    def upgradeObjects(self, calendarName: str):
        """Converts all objects in self.objects to Todos."""
        self.objects = [todo if isinstance(todo, Todo) else Todo.upgrade(todo, calendarName) for todo in self.objects]

    @staticmethod
    def upgrade(synchronisable: caldav.SynchronizableCalendarObjectCollection, calendarName: str) -> "Syncable":
        """Upgrades the third-party caldav.SynchronizableCalendarObjectCollection to a Syncable

        Args:
            synchronisable (caldav.SynchronizableCalendarObjectCollection): the original instance

        Returns:
            (Syncable): the syncable
        """
        syncable = Syncable(synchronisable.calendar, synchronisable.objects, synchronisable.sync_token)  # type: ignore
        syncable.upgradeObjects(calendarName)
        return syncable



"""In this submodule, we define the calendar list widget (on the left)."""
from typing import Iterable

from PySide6 import QtWidgets
from PySide6.QtCore import QModelIndex, QPersistentModelIndex, QSize, Qt
from PySide6.QtGui import QIcon

from melon.calendar import Calendar


class LargerListViewDelegate(QtWidgets.QItemDelegate):
    """An item delegate that slightly increases the display size of each item."""

    def sizeHint(self, option: QtWidgets.QStyleOptionViewItem, index: QModelIndex | QPersistentModelIndex) -> QSize:
        """Returns a size hint for each list widget item, constant in this case.

        Args:
            option (QStyleOptionViewItem): Argument
            index (Union[QModelIndex, QPersistentModelIndex]): Argument

        Returns:
            (QSize):
        """
        return QSize(100, 27)


class CalendarListView(QtWidgets.QListWidget):
    """QListWidget subclass that shows the list of calendars."""

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        """
        Args:
            parent (Union[QWidget, None], optional): Argument
                (default is None)

        """
        super().__init__(parent)
        self.setItemDelegate(LargerListViewDelegate())
        policy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Expanding)
        policy.setHorizontalStretch(2)
        self.setSizePolicy(policy)

        homeItem = QtWidgets.QListWidgetItem(QIcon.fromTheme("go-home"), "All Tasks")
        homeItem.setData(Qt.ItemDataRole.UserRole, {"is-special": True, "specialty": "all"})
        self.addItem(homeItem)

    def populate(self, calendars: Iterable[Calendar]):
        """
        Args:
            calendars (Iterable[Calendar]): Argument
        """
        icon = QIcon.fromTheme("view-list-symbolic")
        for calendar in calendars:
            assert calendar.name is not None
            item = QtWidgets.QListWidgetItem(icon, calendar.name)
            self.addItem(item)
        self.sortItems()



"""This module defines how tasks should be rendered in the GUI."""
import datetime
from typing import Callable

import dateparser.search
from PySide6 import QtWidgets
from PySide6.QtCore import QMargins, QModelIndex, QPersistentModelIndex, QRect, QSize, Qt, QThreadPool, QTimer, Signal
from PySide6.QtGui import QColor, QFont, QPainter, QPainterPath, QPen

from melon.todo import Todo

from .taskwidgets import UserRole

ONE_DAY = datetime.timedelta(days=1)


class TaskItemEditorFactory(QtWidgets.QItemEditorFactory):
    """Factory for task item *editors*."""

    def __init__(self) -> None:
        """Initialise the editor factory, creating a ThreadPool"""
        super().__init__()
        self.threadPool = QThreadPool()

    def createEditor(self, userType: int, parent: QtWidgets.QWidget) -> QtWidgets.QWidget:
        """
        Args:
            userType (int): Argument
            parent (QWidget): Argument

        Returns:
            (QWidget):
        """
        edit = QtWidgets.QLineEdit(parent)
        edit.setAlignment(Qt.AlignmentFlag.AlignTop)
        edit.setContentsMargins(18 + 32 + 10, 2, 2, 4)
        label = QtWidgets.QLabel("Adding task...", edit)
        label.setFixedWidth(245)
        label.move(parent.width() - 250, 22)
        label.setAlignment(Qt.AlignmentFlag.AlignRight)
        edit.textChanged.connect(self.textChangeHandler(edit, label))
        return edit

    def textChangeHandler(self, edit: QtWidgets.QLineEdit, label: QtWidgets.QLabel) -> Callable:
        """Returns a handler function for keypress events on the editor.
        The handler interacts with a timer to prevent computing too frequently.

        Args:
            edit (QtWidgets.QLineEdit): the line edit widget whose events we are catching
            label (QtWidgets.QLabel): the label to display information in

        Returns:
            Callable: the event handler
        """

        def actualHandler():
            """I am called to perform the actual parsing."""
            text = edit.text()
            results = dateparser.search.search_dates(text)
            if results:
                token, stamp = results[0]
                label.setText(stamp.strftime("%d.%m.%Y"))
            else:
                label.clear()

        timer = QTimer()
        timer.setSingleShot(True)
        timer.setInterval(800)
        timer.timeout.connect(lambda: self.threadPool.start(actualHandler))

        def handler():
            """I am called on each keystroke of the QLineEdit."""
            timer.start()

        return handler


class TaskItemDelegate(QtWidgets.QStyledItemDelegate):
    """The task item delegate responsible for rendering todos (= tasks)."""

    editorDestroyed = Signal(QModelIndex)

    def __init__(self, parent):
        """
        Args:
            parent (QObject | None, optional): Argument
                (default is None)
        """
        super().__init__(parent)
        self.setItemEditorFactory(TaskItemEditorFactory())

    def destroyEditor(self, editor: QtWidgets.QWidget, index: QModelIndex | QPersistentModelIndex) -> None:
        """
        Args:
            editor (QWidget): Argument
            index (Union[QModelIndex, QPersistentModelIndex]): Argument
        """
        super().destroyEditor(editor, index)
        self.editorDestroyed.emit(index)

    def paint(
        self,
        painter: QPainter,
        option: QtWidgets.QStyleOptionViewItem,
        index: QModelIndex | QPersistentModelIndex,
    ):
        """
        Args:
            painter (QPainter): Argument
            option (QStyleOptionViewItem): Argument
            index (Union[QModelIndex, QPersistentModelIndex]): Argument
        """
        todo: Todo = index.data(UserRole)
        if todo is None:
            return

        rect: QRect = option.rect  # type: ignore
        painter.save()
        if todo.priority < 3:
            painter.setPen(QColor(255, 100, 100))
        elif todo.priority < 7:
            painter.setPen(QColor(33, 150, 243))
        painter.drawText(rect.translated(18 + 32 + 14, 3), todo.summary)

        dueDate = todo.dueDate
        if dueDate:
            dueTime = todo.dueTime
            text = dueDate.strftime("%d.%m.%Y")
            today = datetime.date.today()
            if dueDate == today:
                text = "today"
            elif dueDate == today - ONE_DAY:
                text = "yesterday"
            elif dueDate == today + ONE_DAY:
                text = "tomorrow"
            painter.setPen(QPen(QColor(255, 100, 100) if dueDate < today else QColor(150, 150, 150)))
            if dueTime is not None:
                text += ", " + dueTime.strftime("%H:%M")
            painter.drawText(rect.translated(-10, 3), text, Qt.AlignmentFlag.AlignRight)

        path = QPainterPath()
        path.addRoundedRect(rect.marginsRemoved(QMargins(2, 2, 2, 4)), 6, 6)
        painter.setPen(QPen(QColor(0, 255, 0, 150)))
        painter.fillPath(path, QColor(200, 200, 200, 30))
        if option.state & QtWidgets.QStyle.StateFlag.State_Selected:  # type: ignore
            path = QPainterPath()
            path.addRoundedRect(QRect(rect.x() + 2, rect.y() + 2, 8, 44), 6, 6)
            painter.fillPath(path, QColor(33, 150, 243, 200))

        if todo.calendarName is not None:
            path = QPainterPath()
            path.addRoundedRect(QRect(rect.x() + 60, rect.y() + 25, len(todo.calendarName) * 8 + 12, 16), 10, 10)
            painter.drawPath(path)
            painter.setFont(QFont("Monospace", 9))
            painter.drawText(rect.translated(32 + 14 + 22, 26), todo.calendarName)

        painter.restore()

    def sizeHint(self, option: QtWidgets.QStyleOptionViewItem, index: QModelIndex) -> QSize:
        """
        Args:
            option (QStyleOptionViewItem): Argument
            index (QModelIndex): Argument

        Returns:
            (QSize):
        """
        return QSize(100, 50)



from typing import Iterable

def schedule(tasks: Iterable[tuple[str, float, int, int, float]]) -> Iterable[tuple[str, float, float]]:
    """Schedules the given tasks in low-level representation into calendar.

    Args:
        tasks (Iterable[tuple[str, float, int, int]]): vector of tasks (uid, duration, priority, location)

    Returns:
        Iterable[tuple[str, float, float]]: vector of allocated timeslots (uid, timestamp, duration)
    """



"""A collection of (quality measure) visualisation helpers."""
from typing import Sequence

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np


def radarChart(values: tuple[float, float, float], title: str, filename: str | None = None):
    """Plots a helpful priority chart.

    Adapted from https://gist.github.com/sausheong/3997c7ba8f42278866d2d15f9e63f7ad.

    Args:
        data (tuple[float, float, float]): data
        title (str): titles of plots
    """
    labels = ["Completed Tasks", "Important Tasks", "In Time Tasks"]
    angles = np.linspace(0, 2 * np.pi, len(labels), endpoint=False)

    # the following three lines are only here to close the polygon
    data = values + (values[0],)
    labels.append(labels[0])
    angles = np.concatenate((angles, [angles[0]]))

    fig = plt.figure()
    axes = fig.add_subplot(1, 1, 1, polar=True)
    axes.set_title(title, weight="bold", size="large")
    axes.plot(angles, data, "o-", linewidth=2, color="blueviolet")
    axes.fill(angles, data, alpha=0.25, color="blueviolet")
    axes.set_thetagrids(angles * 180 / np.pi, labels)
    axes.set_ylim(0, 1.0)
    axes.grid(True)
    if filename is not None:
        fig.savefig(filename)  # type: ignore


def plotConvergence(data: np.ndarray, labels: Sequence, filename: str | None = None):
    """Plots convergence data to a file

    Args:
        data (np.array): data of temp, E_avg, E_var
        filename (str): path to file
    """
    fig = plt.figure()
    axes: matplotlib.axes.Axes = fig.add_subplot(3, 1, 1)
    for i in range(data.shape[0]):
        axes.plot(data[i, :, 0], label=labels[i])
    axes.set_xlabel("Iteration")
    axes.set_ylabel("Temperature $T$")
    axes.legend()
    axes: matplotlib.axes.Axes = fig.add_subplot(3, 1, 2)
    for i in range(data.shape[0]):
        axes.plot(data[i, :, 1], label=labels[i])
    axes.set_xlabel("Iteration")
    axes.set_ylabel("$E_{avg}$")
    axes.legend()
    axes: matplotlib.axes.Axes = fig.add_subplot(3, 1, 3)
    for i in range(data.shape[0]):
        axes.plot(data[i, :, 2], label=labels[i])
    axes.set_xlabel("Iteration")
    axes.set_ylabel("$E_{var}$")
    axes.legend()
    if filename is not None:
        fig.savefig(filename)  # type: ignore



"""The scheduler algorithm"""
import logging
import math
import random
from datetime import date, datetime, timedelta
from typing import Iterable, Mapping

from .base import DAY_LENGTH, INITIAL_TEMPERATURE, START_OF_DAY, SWEEP_EXPONENT, AbstractScheduler, Task, TimeSlot


class AvailabilityManager:
    """This class manages the user's availability in a calendar."""

    def __init__(self) -> None:
        """Initialises the availability manager according to defaults."""
        self.startOfDay = START_OF_DAY  # start at 10am
        self.defaultDayLength = DAY_LENGTH  # going all the way to 2am

    def startingSlot(self) -> TimeSlot:
        """Starting slot, starting at 10am today

        Returns:
            TimeSlot: the first working slot
        """
        return TimeSlot(datetime.combine(date.today(), self.startOfDay), self.defaultDayLength)

    def generateNextSlot(self, previous: TimeSlot) -> TimeSlot:
        """Following a daily schedule, returns the next possible working slot

        Args:
            previous (TimeSlot): the previous working slot

        Returns:
            TimeSlot: the next working slot
        """
        return TimeSlot(
            datetime.combine(previous.timestamp.date() + timedelta(days=1), self.startOfDay),
            self.defaultDayLength,
        )

    def spreadTasks(self, tasks: Iterable[Task]) -> Iterable[tuple[str, TimeSlot]]:
        """Spreads the given list of tasks across the available slots in the calendar, in order.

        Args:
            tasks (Iterable[Task]): list of tasks to schedule

        Yields:
            Iterator[tuple[str, TimeSlot]]: pairs of (UID, TimeSlot), returned in chronological order
        """
        slot = self.startingSlot()
        stamp = slot.timestamp
        for task in tasks:
            if task.duration > self.defaultDayLength:
                raise RuntimeError(
                    "You are trying to schedule a task longer than any working slot."
                    "Split it into smaller chunks! Automatic splitting is not supported."
                )
            taskDuration = timedelta(hours=task.duration)
            if stamp + taskDuration > slot.end:
                slot = self.generateNextSlot(slot)
                stamp = slot.timestamp
            yield (task.uid, TimeSlot(stamp, taskDuration.total_seconds() / 3600))
            stamp += taskDuration


class MCMCScheduler(AbstractScheduler):
    """MCMC class to schedule tasks to events in a calendar."""

    State = tuple[int, ...]  # using literal ellipsis to indicate a homogenous tuple of ints

    def __init__(self, tasks: list[Task]) -> None:
        """Initialises the MCMC scheduler, working on a set of pre-defined tasks.

        Args:
            tasks (list[Task]): the tasks to be scheduled
        """
        super().__init__(tasks)
        self.availability = AvailabilityManager()
        self.state = tuple(range(len(self.tasks)))  # initialise in order
        self.temperature = 1.0
        self.energyLog = []
        self.sweepExponent = SWEEP_EXPONENT
        self.constantEnergyMinimum = sum(t.duration for t in self.tasks) + sum((n) for n in self.state)

    def permuteState(self) -> State:
        """Proposes a new state to use instead of the old state.

        Returns:
            State: the new state, a list of indices within self.tasks representing traversal order
        """
        newState = list(self.state)
        indexA = random.randrange(len(newState))
        indexB = random.randrange(len(newState))
        indexAValue = self.state[indexA]
        newState[indexA] = self.state[indexB]
        newState[indexB] = indexAValue
        return tuple(newState)

    def computeEnergy(self, state: State) -> float:
        """For the given state, compute an MCMC energy (the lower, the better)

        Args:
            state (State): state of the MCMC algorithm

        Returns:
            float: the energy / penalty for this state
        """
        spread = list(self.availability.spreadTasks(self.tasks[i] for i in state))
        totalTimePenalty = (spread[-1][1].end - spread[0][1].timestamp).total_seconds() / 3600
        priorityPenalty = sum((position) * self.tasks[state[position]].priority for position in range(len(state)))
        commutePenalty = 0.0
        onTimePenalty = 0.0
        for position in range(1, len(state)):
            previous = self.tasks[state[position - 1]]
            current = self.tasks[state[position]]
            if current.due is not None and current.due < spread[position][1].end:
                onTimePenalty += 100.0
            if previous.location == 0 or current.location == 0:
                continue  # hybrid tasks can be done from anywhere, so do not penalise
            if previous.location != current.location:
                commutePenalty += 30.0
        total = totalTimePenalty + priorityPenalty + commutePenalty + onTimePenalty - self.constantEnergyMinimum
        return total

    def mcmcSweep(self):
        """Performs a full MCMC sweep"""
        energy = self.computeEnergy(self.state)
        E_sum, E_squared_sum = 0, 0
        steps = len(self.tasks) ** 2
        for i in range(steps):
            newState = self.permuteState()
            delta = self.computeEnergy(newState) - energy
            acceptanceProbability = min(math.exp(-delta / (energy * self.temperature)), 1)
            if random.random() < acceptanceProbability:
                self.state = newState
                energy += delta
            E_sum += energy
            E_squared_sum += energy**2
        E_avg = E_sum / steps
        E_var = E_squared_sum / steps - E_avg**2
        self.energyLog.append((self.temperature, E_avg, E_var))

    def schedule(self) -> Mapping[str, TimeSlot]:
        """Schedules the tasks using an MCMC procedure.

        Returns:
            Mapping[str, TimeSlot]: the resulting map of Tasks to TimeSlots
        """
        for k in range(1, 16):
            self.temperature = INITIAL_TEMPERATURE * k**self.sweepExponent
            self.mcmcSweep()
        logging.info(f"Final State of the MCMC simulation {self.state}.")
        return dict(self.availability.spreadTasks(self.tasks[i] for i in self.state))



"""This module defines the task list view widget."""
import logging
import re

import dateparser.search
from PySide6 import QtWidgets
from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon, QKeyEvent, QMouseEvent

from melon.melon import Melon
from melon.todo import Todo

from .taskitemdelegate import TaskItemDelegate
from .taskwidgets import OrderableTaskItem, TaskOverlayWidget, UserRole


class TaskListView(QtWidgets.QListWidget):
    """Subclass of QListWidget containing tasks."""

    def __init__(self, melon: Melon):
        """Initialise the task list, a QListWidget displaying tasks in a list, using data from `melon`.

        Args:
            melon (Melon): The Melon instance to obtain data from
        """
        super().__init__()
        self.melon = melon
        delegate = TaskItemDelegate(self)
        self.setItemDelegate(delegate)
        delegate.editorDestroyed.connect(self.delegateEditorDestroyed)
        self.setDragEnabled(True)
        self.setVerticalScrollMode(QtWidgets.QListWidget.ScrollMode.ScrollPerPixel)
        self.itemChanged.connect(self.onItemChange)
        self._currentCalendarName = None
        self.addAddButton()

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        """Handles mouse double click

        Args:
            event (QMouseEvent): Argument
        """
        item = self.itemAt(event.position().toPoint())
        self.removeItemWidget(item)
        self.editItem(item)

    def addTask(self, task: Todo) -> OrderableTaskItem:
        """Adds a Todo to the list

        Args:
            task (Todo): Argument

        Returns:
            (MyListWidgetItem):
        """
        item = OrderableTaskItem(task.summary)
        item.setData(UserRole, task)
        item.setFlags(item.flags() | Qt.ItemFlag.ItemIsEditable | Qt.ItemFlag.ItemIsDragEnabled)
        self.addItem(item)
        self.attachTaskWidget(item)
        return item

    def attachTaskWidget(self, item):
        """For a given ListWidgetItem, attaches a task widget

        Args:
            item: Argument
        """
        widget = TaskOverlayWidget(parent=self)
        widget.completionBtn.clicked.connect(lambda: self.completeTask(item))
        self.setItemWidget(item, widget)

    def completeTask(self, item: QtWidgets.QListWidgetItem):
        """Completes the task associated with the given item.

        Args:
            item (QListWidgetItem): Argument
        """
        task: Todo = item.data(UserRole)
        task.complete()
        self.takeItem(self.row(item))

    def addAddButton(self):
        """Adds the task creation button at the bottom of the list."""
        self._addTaskItem = OrderableTaskItem("Add Task")
        self._addTaskItem.setData(Qt.ItemDataRole.EditRole, "add-task")
        self.addItem(self._addTaskItem)
        addButton = QtWidgets.QPushButton(QIcon.fromTheme("list-add"), "Add Task")
        addButton.clicked.connect(self.addEmptyTask)
        self.setItemWidget(self._addTaskItem, addButton)

    def setCalendarFilter(self, calendarName):
        """Only shows tasks for the given calendar.

        Args:
            calendarName: Argument
        """
        for i in range(self.count()):
            item = self.item(i)
            if calendarName is not None and item.data(UserRole) is not None:
                item.setHidden(item.data(UserRole).calendarName != calendarName)
            else:
                item.setHidden(False)
        self.sortItems()
        self._currentCalendarName = calendarName

    def clearCalendarFilter(self):
        """Clears the calendar filter, hence shows all tasks"""
        for i in range(self.count()):
            self.item(i).setHidden(False)
        self._currentCalendarName = None

    def onItemChange(self, item: QtWidgets.QListWidgetItem):
        """Called when an item's value changes. Handles saving of the underlying object.

        Args:
            item (QListWidgetItem): Argument
        """
        todo: Todo = item.data(UserRole)
        if todo.calendarName is None:
            raise ValueError("The item's associated todo does not have a calendar.")

        text = item.text()
        if "clear" in text:
            todo.dueDate = None
            text = text.replace("clear", "")
        else:
            extractedDateTime = dateparser.search.search_dates(text)
            if extractedDateTime:
                token, stamp = extractedDateTime[0]
                if re.search(r"\d", token):
                    todo.dueDate = stamp
                else:
                    todo.dueDate = stamp.date()
        todo.summary = text.strip()
        todo.save()
        logging.info("... saved!")
        self.melon.syncCalendar(self.melon.calendars[todo.calendarName])
        logging.debug("... and synced!")

    def addEmptyTask(self):
        """Add an empty task to the bottom for editing and saving later."""
        if self._currentCalendarName is None:
            logging.warning("Please select a calendar first!")
            return
        calendar = self.melon.calendars[self._currentCalendarName]
        todo = calendar.createTodo()
        item = self.addTask(todo)
        self.sortItems()
        self.removeItemWidget(item)
        self.scrollToItem(item)
        self.editItem(item)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """Handles keypresses.

        Args:
            event (QKeyEvent): Argument
        """
        return super().keyPressEvent(event)

    def delegateEditorDestroyed(self, index):
        """When the editor is destroyed, re-attach the item widget

        Args:
            index: Argument
        """
        if self.itemWidget(self.itemFromIndex(index)) is None:
            self.attachTaskWidget(self.itemFromIndex(index))



"""The scheduler algorithm"""
import dataclasses
import random
from datetime import datetime, time, timedelta
from typing import Mapping

START_OF_DAY = time(10, 0)
DAY_LENGTH = 14
INITIAL_TEMPERATURE = 0.4
SWEEP_EXPONENT = -2.0


@dataclasses.dataclass
class Task:
    """Slim struct representing a task"""

    uid: str  # unique identifier of the task
    duration: float  # estimated, in hours
    priority: int  # between 1 and 9
    location: int  # number indicating the location, where 0 is "hybrid"
    due: datetime | None  # when the task is due

    def asTuple(self, start: datetime) -> tuple[str, float, int, int, float]:
        """Returns a low-level representation of this instance.

        Args:
            start (datetime): Start time reference for the due date

        Returns:
            tuple[str, float, int, int, float]: low-level representation (uid, duration, priority, location, due).
                due is 0 if there is no due date.
        """
        return (
            self.uid,
            self.duration,
            self.priority,
            self.location,
            (self.due - start).total_seconds() / 3600 if self.due is not None else 0,
        )


@dataclasses.dataclass
class TimeSlot:
    """Slim struct representing a time slot, so an event consisting of a start and end date."""

    timestamp: datetime
    duration: float  # in hours

    @property
    def timedelta(self) -> timedelta:
        """
        Returns:
            timedelta: the duration as a datetime.timedelta instance
        """
        return timedelta(hours=self.duration)

    @property
    def end(self) -> datetime:
        """
        Returns:
            datetime: the end timestamp of this time slot
        """
        return self.timestamp + self.timedelta


class AbstractScheduler:
    """Abstract Base Class (ABC) for schedulers."""

    def __init__(self, tasks: list[Task]) -> None:
        """Initialises the scheduler, working on a set of pre-defined tasks.

        Args:
            tasks (list[Task]): the tasks to be scheduled
        """
        self.tasks = tasks

    def uidTaskMap(self) -> Mapping[str, Task]:
        """Generates a dictionary for task lookup by UID.

        Returns:
            Mapping[str, Task]: the dictionary keyed by UID of each task.
        """
        return {task.uid: task for task in self.tasks}

    def schedule(self) -> Mapping[str, TimeSlot]:
        """Schedules the tasks using an MCMC procedure.

        Returns:
            Mapping[str, TimeSlot]: the resulting map of Tasks to TimeSlots
        """
        raise NotImplementedError()


def generateDemoTasks() -> list[Task]:
    """Generates a fixed set of demo tasks.

    Returns:
        list[Task]: the generated list of tasks
    """
    now = datetime.now()
    return [
        Task("1", 3.5, 1, 1, now + timedelta(hours=3)),
        Task("2", 2.0, 7, 2, None),
        Task("3", 11.0, 3, 1, now + timedelta(hours=22)),
        Task("4", 2.0, 9, 0, now + timedelta(hours=100)),
        Task("5", 4.0, 5, 1, None),
    ]


def generateManyDemoTasks(N: int, proportionOfDueDates: float = 0.5) -> list[Task]:
    """Generates a larger set of randomly generated demo tasks.

    Args:
        N (int): Number of tasks to be generated
        proportionOfDueDates (float, optional): what percentage (from 0 to 1) of tasks should have a due date.
                                                Defaults to 0.5.

    Returns:
        list[Task]: the list of tasks
    """
    now = datetime.now()
    return [
        Task(
            uid=str(i),
            duration=random.randint(1, 2 * DAY_LENGTH) / 2,
            priority=random.randint(1, 9),
            location=random.randint(0, 2),
            due=now + timedelta(hours=random.randint(10, N * 5)) if random.random() < proportionOfDueDates else None,
        )
        for i in range(N)
    ]



"""Numba implementation of the scheduler algorithm."""
import math
import random
from datetime import date, datetime, timedelta
from typing import Mapping, Sequence

import numba
from numba.typed.typedlist import List as NumbaList

from .base import DAY_LENGTH, INITIAL_TEMPERATURE, START_OF_DAY, SWEEP_EXPONENT, AbstractScheduler, TimeSlot

State = list[int]


@numba.njit()
def spreadTasks(tasks: Sequence[tuple[str, float, int, int, float]]) -> Sequence[tuple[str, float, float]]:
    """Spreads the given list of tasks across the available slots in the calendar, in order.

    Args:
        tasks (Sequence[Task]): list of tasks to schedule

    Yields:
        Iterator[tuple[str, TimeSlot]]: pairs of (UID, TimeSlot), returned in chronological order
    """
    order = []
    slot = (0.0, DAY_LENGTH)
    stamp = slot[0]
    for task in tasks:
        if task[1] > DAY_LENGTH:
            raise RuntimeError(
                "You are trying to schedule a task longer than any working slot."
                "Split it into smaller chunks! Automatic splitting is not supported."
            )
        if stamp + task[1] > slot[0] + slot[1]:
            slot = (slot[0] + 24.0, DAY_LENGTH)
            stamp = slot[0]
        order.append((task[0], stamp, task[1]))
        stamp += task[1]
    return order


@numba.njit()
def permuteState(state: State) -> State:
    """Proposes a new state to use instead of the old state.

    Returns:
        State: the new state, a list of indices within tasks representing traversal order
    """
    newState = state.copy()
    indexA = random.randrange(len(newState))
    indexB = random.randrange(len(newState))
    indexAValue = state[indexA]
    newState[indexA] = state[indexB]
    newState[indexB] = indexAValue
    return newState


@numba.njit()
def computeEnergy(tasks: Sequence[tuple[str, float, int, int, float]], state: State) -> float:
    """For the given state, compute an MCMC energy (the lower, the better)

    Args:
        tasks (Sequence[tuple[str, float, int, int, float]]): list of tasks (uid, duration, priority, location, due)
        state (State): state of the MCMC algorithm

    Returns:
        float: the energy / penalty for this state
    """
    spread = spreadTasks([tasks[i] for i in state])
    totalTimePenalty = spread[-1][1] + spread[-1][2] - spread[0][1]
    priorityPenalty = sum([position * tasks[state[position]][2] for position in range(len(state))])
    commutePenalty = 0.0
    onTimePenalty = 0.0
    for position in range(1, len(state)):
        previous = tasks[state[position - 1]]
        current = tasks[state[position]]
        if current[4] != 0 and current[4] < spread[position][1] + spread[position][2]:
            onTimePenalty += 100
        if previous[3] == 0 or current[3] == 0:
            continue  # hybrid tasks can be done from anywhere, so do not penalise
        if previous[3] != current[3]:
            commutePenalty += 30.0
    # print(totalTimePenalty, priorityPenalty, commutePenalty, onTimePenalty)
    return totalTimePenalty + priorityPenalty + commutePenalty + onTimePenalty


@numba.njit()
def mcmcSweep(tasks: Sequence[tuple[str, float, int, int, float]], initialState: State, temperature: float) -> State:
    """Performs a full MCMC sweep

    Args:
        tasks (Sequence[tuple[str, float, int, int, float]]): list of tasks
        initialState (State): initial ordering
        temperature (float): temperature for Simulated Annealing

    Returns:
        State: new state
    """
    state = initialState
    energy = computeEnergy(tasks, state)
    for i in range(len(tasks) ** 2):
        newState = permuteState(state)
        delta = computeEnergy(tasks, newState) - energy
        acceptanceProbability = min(math.exp(-delta / (energy * temperature)), 1)
        # print(f"New state with energy {energy + delta} (delta {delta}), accepted with {acceptanceProbability}.")
        if random.random() < acceptanceProbability:
            state = newState
            energy += delta
    return state


@numba.njit()
def schedule(tasks: Sequence[tuple[str, float, int, int, float]]) -> Sequence[tuple[str, float, float]]:
    """Schedules the given tasks in low-level representation into calendar.

    Args:
        tasks (Sequence[tuple[str, float, int, int, float]]): vector of tasks (uid, duration, priority, location, due)

    Returns:
        Sequence[tuple[str, float, float]]: vector of allocated timeslots (uid, timestamp, duration)
    """
    state = list(range(len(tasks)))
    for k in range(1, 16):
        temperature = INITIAL_TEMPERATURE * k**SWEEP_EXPONENT
        state = mcmcSweep(tasks, state, temperature)
    return spreadTasks([tasks[i] for i in state])


class NumbaMCMCScheduler(AbstractScheduler):
    """Markov Chain Monte-Carlo Task Scheduler, implemented in Python with numba speed-up."""

    def schedule(self) -> Mapping[str, TimeSlot]:
        """Runs the Rust implementation of the scheduler.

        Returns:
            Mapping[str, TimeSlot]: the resulting schedule
        """
        start = datetime.combine(date.today(), START_OF_DAY)  # equivalent to t = 0
        result = schedule(NumbaList(task.asTuple(start) for task in self.tasks))
        return {t[0]: TimeSlot(start + timedelta(hours=t[1]), t[2]) for t in result}



"""A collection of widgets and classes for the definition of task display in the UI."""
import pathlib

from PySide6 import QtWidgets
from PySide6.QtCore import QRect, Qt
from PySide6.QtGui import QIcon, QPainter, QPaintEvent

from melon.todo import Todo

ASSET_FOLDER = pathlib.Path(__file__).resolve().parent / "assets"
ADD_TASK_EDIT_ROLE = "add-task"
UserRole = Qt.ItemDataRole.UserRole


class OrderableTaskItem(QtWidgets.QListWidgetItem):
    """A subclass of QListWidgetItem that can be sorted according to a few specific rules, i.e. due date."""

    def __lt__(self, other: QtWidgets.QListWidgetItem):
        """This method will by called by QListWidget.sort() to compare items to one another.

        Args:
            other (QListWidgetItem): Argument
        """
        if self.data(Qt.ItemDataRole.EditRole) == ADD_TASK_EDIT_ROLE:
            return False
        elif other.data(Qt.ItemDataRole.EditRole) == ADD_TASK_EDIT_ROLE:
            return True
        mine: Todo = self.data(UserRole)
        theirs: Todo = other.data(UserRole)
        return mine < theirs


class CompletionPushButton(QtWidgets.QPushButton):
    """A push button that only displays a specific icon."""

    def __init__(self, parent: QtWidgets.QWidget):
        """
        Args:
            parent (QWidget): Argument
        """
        super().__init__(parent=parent)
        self.okIcon = QIcon(str(ASSET_FOLDER / "complete.png"))
        self.setFixedSize(34, 34)

    def paintEvent(self, event: QPaintEvent) -> None:
        """
        Args:
            event (QPaintEvent): Argument
        """
        painter = QPainter(self)
        delta = 2 if self.isDown() else 0
        self.okIcon.paint(painter, QRect(delta, delta, 32, 32))


class TaskOverlayWidget(QtWidgets.QWidget):
    """This is the task overlay widget we put above todos in the list display."""

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        """Initialises the widget and adds the necessary GUI subcomponents."""
        super().__init__(parent)
        self.completionBtn = CompletionPushButton(parent=self)
        self.completionBtn.move(18, 8)



"""This submodule only does one thing: loading configuration from the right place."""
import pathlib

try:
    import tomllib
except ImportError:  # pragma: no cover
    import tomli as tomllib  # pragma: no cover

CONFIG_FOLDER = pathlib.Path.home() / ".config" / "melon"
CONFIG_FOLDER.mkdir(exist_ok=True)
CONFIG_PATH = CONFIG_FOLDER / "config.toml"

CONFIG = {"client": {"url": "http://localhost:8000/dav/user/calendars/", "username": None, "password": None}}


def load_config():
    """Loads, or re-loads, the configuration file."""
    global CONFIG
    if CONFIG_PATH.exists():
        with open(CONFIG_PATH, "rb") as f:
            CONFIG = tomllib.load(f)


load_config()



"""Index file of the Graphical User Interface shipped next to melon, called Melon UI.
In order to start the gui, please follow the standard Qt procedure:

from PySide6 import QtWidgets
from melongui.mainwindow import MainWindow

if __name__ == "__main__":
    app = QtWidgets.QApplication()
    window = MainWindow()
    window.buildUI()
    window.start()
    window.show()
    sys.exit(app.exec())
"""



import sys

from PySide6 import QtWidgets
from PySide6.QtCore import QSize

from melongui.mainwindow import MainWindow


def main():
    """Starts the Graphical User Interface"""
    app = QtWidgets.QApplication(sys.argv)
    app.setApplicationName("Melon")
    window = MainWindow()
    window.buildUI()
    window.start()
    window.show()
    screenSize: QSize = app.primaryScreen().availableSize()
    window.resize(screenSize.width() // 2, screenSize.height())
    window.move(screenSize.width() // 2, 0)
    sys.exit(app.exec())
